import java.awt.*;

/**
   Class for displaying a drawing board.
   @version 1.00 4 Dec 1997
   @author Octavian Cismasu
*/      
public class DrawingBoard extends Canvas {
   // class variables
   private Polygon poly;
   private int[] xCoords={0};
   private int[] yCoords={0};
   private int xmouse=-1, ymouse=-1;
   private PlumblineAlgorithm pcont;
   private BigPoint testp;
   private BigPoint infp;  
    // flags
   private boolean eraseFlag = false;
   private boolean existStartPoint = false;
   private boolean polydone = false;
   private boolean existTestpoint=false;
   // constants 
   private static String msg1 = "Click on white area to choose the first point of the polygon";
   private static String msg2 = "Choose a point for inclusion testing";
   private static String msg3 = "When finished, click on 'Close Poly'";
   private static int MAX_INT_POINTS = 500;
   
   
   /**
      Constructs a new DrawingBoard object that 
      has a handle on its parent container.
      @param parentContainer the parent container of this object
   */   
   public DrawingBoard(PlumblineAlgorithm parentContainer) {
      super();
      pcont = parentContainer;
      setBackground(Color.white);   
   } 
     
   
   /**
      Displays graphics in the context of this object
      @param g the graphics context
   */   
   public void paint(Graphics g) {                  
      g.setColor(Color.black);     
      g.setPaintMode();
      g.translate(0,0); // graphics context origin is the (0,0) point
      
      if (eraseFlag) {  // if to be erased    
         g.clearRect(0, 0, size().width, size().height);
         eraseFlag = false;
      }   
      else if (existStartPoint) { // if there is a staring point
         g.drawPolygon(poly);     // draw polyline
         if (!polydone) {         // if polygon is not finished
            if (xmouse != -1)    
               g.drawLine(poly.xpoints[poly.npoints-1], // draw line to current mouse position
                          poly.ypoints[poly.npoints-1], xmouse, ymouse); 
         }         
         else if (existTestpoint) { // finished with polygon 
            g.setColor(Color.blue); 
            testp.drawBigpoint(g);  // show testpoint
            infp.drawBigpoint(g);  
            g.translate(testp.x, testp.y); // graphics context origin is the testpoint
            g.drawLine(0, 0, infp.x, infp.y); // draw ray
            
            // handle intersection points                        
            BigPoint[] cross = inPoly();
            for (int i=0; cross[i]!=null; i++) {               
               if ((i % 2) == 1) 
                  g.setColor(Color.red);
               else 
                  g.setColor(new Color(0,128,0));
               cross[i].drawBigpoint(g);  // show intesection                                   
            }                 
         }
      }                          
   }                                    
     
     
   /**
      Handles mouse events.
      @param evt type of event
      @param x the x-coordinate of the mouse
      @param y the y coordinate of the mouse
   */                                       
   public boolean mouseDown(Event evt, int x, int y) {    
      boolean status = false;
      
      xmouse = ymouse = -1;
      if (!polydone) {   // polygon not finished
         if (!existStartPoint)  // polygon start point does not exist
            startPoint(x,y);    // begin construction of polygon     
         else {
            if (poly.npoints >= 2) {  // polygon has more than 2 points
               pcont.setEnableButtons(true, true, true); // enable all buttons
               Point newp = new Point(x,y); // new point
               Point prevp = new Point(poly.xpoints[poly.npoints-1],
                                       poly.ypoints[poly.npoints-1]); // previous point
               for (int i=1; i<poly.npoints-1; i++){
                  // test for intersection new segment with all other segments
                  Point p1 = new Point(poly.xpoints[i],poly.ypoints[i]);
                  Point p2 = new Point(poly.xpoints[i-1],poly.ypoints[i-1]);
                  if (intersect(newp,prevp,p1,p2)) { // if intersection
                     status = true; // signal it
                     break;         // stop
                  }
               }                              
            }
            if (!status) // if no intersection            
               poly.addPoint(x,y); // add point to polygon
            repaint();
         }                   
      }
      else { // finished with polygon      
         testp = new BigPoint(x,y,6); // our testpoint
         testp.setFill(true);   
         existTestpoint = true;   // signal we have a testpoint
         
         // for point at 'infinity' y coord.= 0,since coordinates are calculated 
         //    with respect to the testpoint (new graphics origin)
         // the big x coordinate is to make sure ray is long enough
         infp = new BigPoint(size().width+200,0,1);  // 'end' point of 'infinite' ray
         repaint();
      }   
                     
      return true;      
   }        
   
   
   /**
      Handles mouse move events.
      @param evt type of event
      @param x the x-coordinate of the mouse
      @param y the y-coordinate of the mouse
   */   
   public boolean mouseMove(Event evt, int x, int y) {
      if (evt.target instanceof DrawingBoard) {
         pcont.setxyCoords(x,y); // show coordinates in status field
         xmouse = x;   // set globals with current mouse position
         ymouse = y;   
         if ((!polydone) && existStartPoint)
            repaint();
      }      
      return true;
   }
    
   
   /**
      Initializes variables for construction 
      of new polygon.
      @param x the x-coordinate of the first polygon vertex
      @param y the y-coordinate of the first polygon vertex
   */   
   public void startPoint(int x, int y) {
      xCoords[0] = x;
      yCoords[0] = y;
      xmouse = -1;
      ymouse = -1;
      poly = new Polygon(xCoords, yCoords, xCoords.length);
      existStartPoint = true; 
      pcont.setMbMessage(msg3,"");
      pcont.setEnableButtons(true, true, false);
   }   
   
   
   
   /**
      Removes the last polygon vertex that was added.      
   */   
   public void undoLastpoint() {
      polydone = false;      
      if (poly.npoints > 1) {
         poly.npoints--;// decrement number of polygon vertices
         repaint();
      }   
      else
         eraseBoard();                
   }      
   
   
   /**
      Erases the graphics area of the DrawingBoard and 
      resets variables. A new polygon can be constructed now.                                                         
   */   
   public void eraseBoard() {
      xmouse = ymouse = -1;
      existStartPoint = false;
      eraseFlag = true;
      polydone = false; 
      existTestpoint = false;
      repaint();
      pcont.setMbMessage(msg1,"");
      pcont.setEnableButtons(true, false, false);     
   }
   
   
   /**
      Finishes the construction of the polygon
      by adding the starting vertex to the current list 
      of points of the polygon.
   */   
   public void closePolygon() {
      boolean status = false;      
      
      xmouse = ymouse = -1;
      // detect intersection of the last edge to be added with the other edges
      Point newp = new Point(xCoords[0],yCoords[0]);
      Point prevp = new Point(poly.xpoints[poly.npoints-1],poly.ypoints[poly.npoints-1]);
      for (int i=2; i<poly.npoints-1; i++){
         Point p1 = new Point(poly.xpoints[i],poly.ypoints[i]);
         Point p2 = new Point(poly.xpoints[i-1],poly.ypoints[i-1]);
         if (intersect(newp,prevp,p1,p2)) { // if intersection
            status = true;     // signal it
            break;             // stop looking
         }
      }         
      if (status == false) {   // if no intersection 
         poly.addPoint(xCoords[0], yCoords[0]);  // add the vertex
         polydone = true;     
         pcont.setMbMessage(msg2,"");
         pcont.setEnableButtons(true,false,false);
      }
      repaint();  
   }   
                  
   

   /**
      Draws a polygon whose coordinates are preset.
      Sets flags accordingly.
      @param xpoints the x coordinates of each point
      @param ypoints the y coordinates of each point
   */    
   public void setPolygon(int[] xpoints, int[] ypoints) {
      eraseBoard();
      poly = new Polygon(xpoints, ypoints, xpoints.length);
      existStartPoint = true;
      polydone = true;
      eraseFlag = false;
      existTestpoint = false;
      repaint();
      pcont.setEnableButtons(true, false, false);      
      pcont.setMbMessage(msg2,"");      
   }   
            

   
   /**
      Tests whether a point is inside the polygon or outside.
      Calculates intersections of ray from testpoint with
      the polygon. Counts the number of such intersections.
      Returns an array of points corresponding to these intersections.
      (Algorithm idea is from <g>O'Rourke</g>)
   */   
   private BigPoint[] inPoly() {
      int crossings = 0;
      Polygon p = transformPolyCoords();
      BigPoint[] crosspts = new BigPoint[MAX_INT_POINTS];
      String result = new String(); 
                       
      for (int i = 0; i < p.npoints-1; i++) {         
         int i1 = (i + p.npoints-2) % (p.npoints-1);
         if (((p.ypoints[i] > 0) && (p.ypoints[i1] <= 0)) ||
             ((p.ypoints[i1] > 0) && (p.ypoints[i] <= 0))) {
            int x = (int)((p.xpoints[i] * p.ypoints[i1] - p.xpoints[i1] * p.ypoints[i]) / 
                                                 (p.ypoints[i1] - p.ypoints[i]));             
            // crosses ray if stricly positive intersection
            if (x > 0) {
               // intersection is always on ray (current x-axis)               
               crosspts[crossings] = new BigPoint(x, 0, 6);
               crosspts[crossings].setFill(true);
               crossings++;
            }   
         }
      }
      if ((crossings % 2) == 1) // if odd number of crossings
         result = "INSIDE";     // point is inside
      else 
         result = "OUTSIDE";    // else, point is outside
      pcont.setMbMessage(crossings+" intersection point(s)."+" The testpoint is......", result); 
      return crosspts;          
   }                           


   
   /**
      Translates the vertices of the polygon
      with respect to the new origin of the graphics context,
      which is the testpoint. The polygon with translated 
      coordinates is returned. 
   */
   private Polygon transformPolyCoords() {
      Polygon trpoly = new Polygon(poly.xpoints, poly.ypoints, poly.npoints);        
      // testpoint is new origin of graphics context so translate accordingly      
      //trpoly.translate(-testp.x, -testp.y);  // this does not work with JDK 1.0.2
      for (int i=0; i<trpoly.npoints; i++) {
         trpoly.xpoints[i] = trpoly.xpoints[i] - testp.x;
         trpoly.ypoints[i] = trpoly.ypoints[i] - testp.y;
      }            
      return trpoly;   
   }      
      
   
   /* Tests for intersection of two segments.
      (Algorithm idea is from <g>O'Rourke</g>).
      @param p1 first point of first segment
      @param p2 second point of first segment 
      @param p3 first point of second segment
      @param p4 second point of second segment
   */   
   private boolean intersect(Point p1,Point p2,Point p3,Point p4) {
      if (intersectProp(p1,p2,p3,p4))
         return true;
      else if (between(p1,p2,p3) || between(p1,p2,p4) || between(p3,p4,p1) || 
         between(p3,p4,p2))
         return true;
      else return false;
   }


   boolean intersectProp(Point p1,Point p2,Point p3,Point p4) {
      if (collinear(p1,p2,p3) || collinear(p1,p2,p4) || collinear(p3,p4,p1) || 
            collinear(p3,p4,p2))
         return false;
      return Xor(Left(p1,p2,p3),Left(p1,p2,p4)) && 
             Xor(Left(p3,p4,p1),Left(p3,p4,p2));
   }

   boolean between(Point p1,Point p2,Point p3) {
      if (! collinear(p1,p2,p3)) return false;
      if (p1.x != p2.x)
         return ((p1.x <= p3.x) && (p3.x <= p2.x) ||
                ((p1.x >= p3.x) && (p3.x >= p2.x)));
      else return ((p1.y <= p3.y) && (p3.y <= p2.y) ||
                   ((p1.y >= p3.y) && (p3.y >= p2.y)));
   }

   boolean collinear(Point p1, Point p2, Point p3){  
      return Area2(p1,p2,p3) == 0;
   }

   private int Area2(Point p1,Point p2,Point p3) {
      return (p2.x - p1.x) * (p3.y - p1.y) - 
        (p3.x - p1.x) * (p2.y - p1.y);
   }

   private boolean Left(Point p1,Point p2,Point p3){
      return Area2(p1,p2,p3) >0;
   }

   private boolean Xor(boolean a,boolean b){
      if ((a && b) || (!a && !b)) return false;
         return true;
   }   
   
}   


